# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from .....schema_classes import DatasetDeprecationClass
from .....schema_classes import DatasetFieldMappingClass
from .....schema_classes import DatasetFieldProfileClass
from .....schema_classes import DatasetFieldUsageCountsClass
from .....schema_classes import DatasetLineageTypeClass
from .....schema_classes import DatasetProfileClass
from .....schema_classes import DatasetPropertiesClass
from .....schema_classes import DatasetUpstreamLineageClass
from .....schema_classes import DatasetUsageStatisticsClass
from .....schema_classes import DatasetUserUsageCountsClass
from .....schema_classes import EditableDatasetPropertiesClass
from .....schema_classes import FineGrainedLineageClass
from .....schema_classes import FineGrainedLineageDownstreamTypeClass
from .....schema_classes import FineGrainedLineageUpstreamTypeClass
from .....schema_classes import HistogramClass
from .....schema_classes import QuantileClass
from .....schema_classes import UpstreamClass
from .....schema_classes import UpstreamLineageClass
from .....schema_classes import ValueFrequencyClass
from .....schema_classes import ViewPropertiesClass


DatasetDeprecation = DatasetDeprecationClass
DatasetFieldMapping = DatasetFieldMappingClass
DatasetFieldProfile = DatasetFieldProfileClass
DatasetFieldUsageCounts = DatasetFieldUsageCountsClass
DatasetLineageType = DatasetLineageTypeClass
DatasetProfile = DatasetProfileClass
DatasetProperties = DatasetPropertiesClass
DatasetUpstreamLineage = DatasetUpstreamLineageClass
DatasetUsageStatistics = DatasetUsageStatisticsClass
DatasetUserUsageCounts = DatasetUserUsageCountsClass
EditableDatasetProperties = EditableDatasetPropertiesClass
FineGrainedLineage = FineGrainedLineageClass
FineGrainedLineageDownstreamType = FineGrainedLineageDownstreamTypeClass
FineGrainedLineageUpstreamType = FineGrainedLineageUpstreamTypeClass
Histogram = HistogramClass
Quantile = QuantileClass
Upstream = UpstreamClass
UpstreamLineage = UpstreamLineageClass
ValueFrequency = ValueFrequencyClass
ViewProperties = ViewPropertiesClass

# fmt: on
