# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from .....schema_classes import DataHubActorFilterClass
from .....schema_classes import DataHubPolicyInfoClass
from .....schema_classes import DataHubResourceFilterClass


DataHubActorFilter = DataHubActorFilterClass
DataHubPolicyInfo = DataHubPolicyInfoClass
DataHubResourceFilter = DataHubResourceFilterClass

# fmt: on
