{
  "type": "record",
  "name": "Container",
  "namespace": "com.linkedin.pegasus2avro.container",
  "doc": "Link from an asset to its parent container",
  "fields": [
    {
      "name": "container",
      "type": "string",
      "doc": "The parent container of an asset",
      "Relationship": {
        "entityTypes": [
          "container"
        ],
        "name": "IsPartOf"
      },
      "Searchable": {
        "addToFilters": true,
        "fieldName": "container",
        "fieldType": "URN",
        "filterNameOverride": "Container"
      },
      "java": {
        "class": "com.linkedin.pegasus2avro.common.urn.Urn"
      }
    }
  ],
  "Aspect": {
    "name": "container"
  }
}