# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from .....schema_classes import DataHubRetentionConfigClass
from .....schema_classes import RetentionClass
from .....schema_classes import TimeBasedRetentionClass
from .....schema_classes import VersionBasedRetentionClass


DataHubRetentionConfig = DataHubRetentionConfigClass
Retention = RetentionClass
TimeBasedRetention = TimeBasedRetentionClass
VersionBasedRetention = VersionBasedRetentionClass

# fmt: on
